// This header file defines standard types and constants
// that all of my programs tend to rely on
// (C) 1993-1995 Stuart Cheshire <cheshire@cs.stanford.edu>

// ************************************************************
// Basic C definitions

#define local static
#define import extern
#define export
#define until(A) while(!(A))
#define elementsof(A) (sizeof(A)/sizeof((A)[0]))
#define semiconst const

typedef unsigned char  u_char;
typedef unsigned short u_short;
typedef unsigned long  u_long;

typedef u_char  NIBBLE;	// to be interpreted as four bits
typedef u_char  BYTE;
typedef u_short WORD;
typedef u_long  DWORD;
typedef union { DWORD dword[2]; WORD word[4]; BYTE byte[ 8]; } QWORD;
typedef union { DWORD dword[4]; WORD word[8]; BYTE byte[16]; } OCTWORD;

typedef union  {   char c[  4];   long l; }   char4;
typedef union  { u_char c[  4]; u_long l; } u_char4;
typedef struct {   char c[ 32]; }   char32;
typedef struct { u_char c[ 32]; } u_char32;
typedef struct {   char c[ 36]; }   char36;
typedef struct { u_char c[ 36]; } u_char36;
typedef struct {   char c[ 64]; }   char64;
typedef struct { u_char c[ 64]; } u_char64;
typedef struct {   char c[ 68]; }   char68;
typedef struct { u_char c[ 68]; } u_char68;
typedef struct {   char c[256]; }   char256;
typedef struct { u_char c[256]; } u_char256;

// ************************************************************
// 68000 processor related definitions

#ifdef __MC68K__

typedef short InterruptStatus;

#pragma parameter __D0 DisableInterrupts()	// MOVE SR,D0  ORI.W #$0700,SR
local InterruptStatus DisableInterrupts(void) = { 0x40C0, 0x007C, 0x0700 };

#pragma parameter RestoreInterrupts(__D0)	// MOVE D0,SR
local void RestoreInterrupts(InterruptStatus oldval) = { 0x46C0 };

#pragma parameter __D0 GetRegisterD0()	// MOVEA.L D0,D0
local void* GetRegisterD0(void) = { 0x2000 };
#pragma parameter __D0 GetRegisterD1()	// MOVEA.L D1,D0
local void* GetRegisterD1(void) = { 0x2001 };
#pragma parameter __D0 GetRegisterD2()	// MOVEA.L D2,D0
local void* GetRegisterD2(void) = { 0x2002 };
#pragma parameter __D0 GetRegisterD3()	// MOVEA.L D3,D0
local void* GetRegisterD3(void) = { 0x2003 };
#pragma parameter __D0 GetRegisterD4()	// MOVEA.L D4,D0
local void* GetRegisterD4(void) = { 0x2004 };
#pragma parameter __D0 GetRegisterD5()	// MOVEA.L D5,D0
local void* GetRegisterD5(void) = { 0x2005 };
#pragma parameter __D0 GetRegisterD6()	// MOVEA.L D6,D0
local void* GetRegisterD6(void) = { 0x2006 };
#pragma parameter __D0 GetRegisterD7()	// MOVEA.L D7,D0
local void* GetRegisterD7(void) = { 0x2007 };

#pragma parameter SetRegisterD0(__D0)	// MOVEA.L D0,D0
local void SetRegisterD0(void*) = { 0x2000 };
#pragma parameter SetRegisterD1(__D0)	// MOVEA.L D0,D1
local void SetRegisterD1(void*) = { 0x2200 };
#pragma parameter SetRegisterD2(__D0)	// MOVEA.L D0,D2
local void SetRegisterD2(void*) = { 0x2400 };
#pragma parameter SetRegisterD3(__D0)	// MOVEA.L D0,D3
local void SetRegisterD3(void*) = { 0x2600 };
#pragma parameter SetRegisterD4(__D0)	// MOVEA.L D0,D4
local void SetRegisterD4(void*) = { 0x2800 };
#pragma parameter SetRegisterD5(__D0)	// MOVEA.L D0,D5
local void SetRegisterD5(void*) = { 0x2A00 };
#pragma parameter SetRegisterD6(__D0)	// MOVEA.L D0,D6
local void SetRegisterD6(void*) = { 0x2C00 };
#pragma parameter SetRegisterD7(__D0)	// MOVEA.L D0,D7
local void SetRegisterD7(void*) = { 0x2E00 };

#pragma parameter __A0 GetRegisterA0()	// MOVEA.L A0,A0
local void* GetRegisterA0(void) = { 0x2048 };
#pragma parameter __A0 GetRegisterA1()	// MOVEA.L A1,A0
local void* GetRegisterA1(void) = { 0x2049 };
#pragma parameter __A0 GetRegisterA2()	// MOVEA.L A2,A0
local void* GetRegisterA2(void) = { 0x204A };
#pragma parameter __A0 GetRegisterA3()	// MOVEA.L A3,A0
local void* GetRegisterA3(void) = { 0x204B };
#pragma parameter __A0 GetRegisterA4()	// MOVEA.L A4,A0
local void* GetRegisterA4(void) = { 0x204C };
#pragma parameter __A0 GetRegisterA5()	// MOVEA.L A5,A0
local void* GetRegisterA5(void) = { 0x204D };

#pragma parameter SetRegisterA0(__A0)	// MOVEA.L A0,A0
local void SetRegisterA0(void*) = { 0x2048 };
#pragma parameter SetRegisterA1(__A0)	// MOVEA.L A0,A1
local void SetRegisterA1(void*) = { 0x2248 };
#pragma parameter SetRegisterA2(__A0)	// MOVEA.L A0,A2
local void SetRegisterA2(void*) = { 0x2448 };
#pragma parameter SetRegisterA3(__A0)	// MOVEA.L A0,A3
local void SetRegisterA3(void*) = { 0x2648 };
#pragma parameter SetRegisterA4(__A0)	// MOVEA.L A0,A4
local void SetRegisterA4(void*) = { 0x2848 };
#pragma parameter SetRegisterA5(__A0)	// MOVEA.L A0,A5
local void SetRegisterA5(void*) = { 0x2A48 };

#endif

// ************************************************************
// Macintosh OS related definitions

// Serial Driver
// #define baud38400 1 (This is now defined in system headers)
#define baud28800 2
#define SerHShake_V2 14
#define SerialBaudCode(X) ((115200/(X)) - 2)
#define SerResetCode(X) (SerialBaudCode(X) | data8 | noParity | stop10)

// Time Manager
#define TMTaskActiveBit 0x8000
#define TickCountHz 60

// File Manager
#define fsNoCache 0x20

// ************************************************************
// Symantec compiler related definitions

#if THINK_C
#if __option(a4_globals)
#define A4GLOBALS
#endif
#endif

#ifdef A4GLOBALS
#define GLOBREG A4
#define GetGlobalsRegister GetRegisterA4
#define SetGlobalsRegister SetRegisterA4
#else
#define GLOBREG A5
#define GetGlobalsRegister GetRegisterA5
#define SetGlobalsRegister SetRegisterA5
#endif
